/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <malloc.h>

#include "glass.h"

/* Returns an empty component */
GlassComponent *_glassComponentNew(void)
{
  GlassComponent *component = malloc(sizeof(GlassComponent));

  if(component == NULL)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for component\n");
    return NULL;
  }

  _glassComponentSetTransform(component, _glassTransformNew(GLASS_NONE));
  component->list = NULL;
  component->apoint = NULL;

  component->apoint_transform = FALSE;

  component->next = NULL;
  component->child = NULL;

  return component;
}

/* Set a components transform */
void _glassComponentSetTransform(GlassComponent *component, GlassTransform *transform)
{
  component->transform = transform;
}

/* Get a components transform */
GlassTransform *_glassComponentGetTransform(GlassComponent *component)
{
  return component->transform;
}

/* Set a components triangle list */
void _glassComponentSetList(GlassComponent *component, GlassTriangleList *list)
{
  component->list = list;
}

/* Get a components triangle list */
GlassTriangleList *_glassComponentGetList(GlassComponent *component)
{
  return component->list;
}

/* Set a components active point */
void _glassComponentSetActivePoint(GlassComponent *component, GlassActivePoint *apoint)
{
  component->apoint = apoint;
}

/* Get a components triangle list */
GlassActivePoint *_glassComponentGetActivePoint(GlassComponent *component)
{
  return component->apoint;
}

/* Add a sibling to a component */
void _glassComponentAddSibling(GlassComponent *component, GlassComponent *sibling)
{
  GlassComponent *com_ptr = component;

  /* Find the end of the list... */
  while(com_ptr->next != NULL)
    com_ptr = com_ptr->next;

  /* ...and attach */
  com_ptr->next = sibling;

  /* Cap end, just in case */
  sibling->next = NULL;
}

/* Remove a sibling from a component, returning FALSE if sibling not found */
int _glassComponentRemoveSibling(GlassComponent *component, GlassComponent *sibling)
{
  GlassComponent *com_ptr = component;

  /* Find that sibling */
  while(com_ptr && com_ptr->next != sibling)
    com_ptr = com_ptr->next;

  /* If sibling wasn't found, return FALSE */
  if(!com_ptr)
    return FALSE;

  /* Skip out sibling */
  com_ptr->next = sibling->next;

  /* Sibling is now isolated */
  sibling->next = NULL;

  return TRUE;
}

/* Add a child to a component */
void _glassComponentAddChild(GlassComponent *component, GlassComponent *child)
{
  if(component->child)
    _glassComponentAddSibling(component->child, child);
  else
    component->child = child;
}

/* Remove a child from a component, returning FALSE if child not found */
int _glassComponentRemoveChild(GlassComponent *component, GlassComponent *child)
{
  if(!component || !child)
    return FALSE;

  if(component->child == child)
    component->child = child->next;
  else
    return _glassComponentRemoveSibling(component->child, child);

  return TRUE;
}

/* Recursively work out which components are needed for any active points */
int _glassComponentTestActivePointr(GlassComponent *component)
{
  int apoint_transform = FALSE;

  if(component == NULL)
    return FALSE;

  /* If this component has an active point, it's important */
  if(_glassComponentGetActivePoint(component) != NULL)
    apoint_transform = TRUE;

  /* If any of this components children have active points,
     then this component is important */
  if(_glassComponentTestActivePointr(component->child))
    apoint_transform = TRUE;

  component->apoint_transform = apoint_transform;

  /* Try the next component, irrelevant to this component if it has any
     acitve points, but relevant to the parent of this component */
  if(_glassComponentTestActivePointr(component->next))
    apoint_transform = TRUE;

  return apoint_transform;
}

/* Recursively update the active points */
void _glassComponentUpdateActivePointsr(GlassComponent *component)
{
  GlassActivePoint *apoint;

  /* Return if hit the end of a list */
  if(component == NULL)
    return;

  /* Move on to the next component */
  _glassComponentUpdateActivePointsr(component->next);
  
  /* Stop if there are no children of this component with active points */
  if(!component->apoint_transform)
    return;

  glPushMatrix();

  /* Do transform (if it has one) */
  _glassDoTransform(_glassComponentGetTransform(component));

  /* If has an active point, update that */
  if((apoint = _glassComponentGetActivePoint(component)) != NULL)
    _glassActivePointUpdate(apoint);

  /* Move onto next component (that inherits this transform) */
  _glassComponentUpdateActivePointsr(component->child);

  glPopMatrix();
}

/* Free a component */
void _glassFreeComponent(GlassComponent *component)
{
  _glassFreeTransform(component->transform);
  free(component);
}

/* Free a component, and (recursively) its siblings/children */
void _glassFreeComponentr(GlassComponent *component)
{
  if(component == NULL)
    return;

  _glassFreeComponentr(component->next);
  _glassFreeComponentr(component->child);

  _glassFreeComponent(component);
}
