/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "glass.h"

#define FLOAT_ERROR 9999999999.9f /* !!There must be a better way!!!! */

/* Load the object file 'fname', use a library if library != NULL */
GlassObject *_glassLoadObject(GlassLibrary *library, const char *fname)
{
  GlassObject *object = NULL; /* the new object */
  /* Contains a list of components and their indicies */
  GlassComponentList *component_list = NULL;
  /* Contains a tree of the components relating structure */
  GlassStructureTree *structure_tree = NULL;
  /* The data file (containing the glass object) */
  FILE *glass_data_file;

  /* The data file */
  glass_data_file = fopen(fname, "r");
  if(!glass_data_file)
  {
    fprintf(stderr, "Glass Error: Unable to open file \"%s\"\n", fname);
    return NULL;
  }

  /* The new object */
  if(!(object = _glassObjectNew("")))
    return NULL;

  if(library)
    _glassLibraryAddObject(library, object);

  /* Go through the file a few times, so loading in the correct order */
  /* !!note: This isn't terribly efficient, but makes the code easier to read */
  _glassLoadObjectInfo(object, glass_data_file, fname);
  _glassLoadTextures(object, glass_data_file, fname);
  _glassLoadMaterials(object, glass_data_file, fname);
  _glassLoadTriangleLists(object, glass_data_file, fname);
  _glassLoadVariables(object, glass_data_file, fname);
  _glassLoadActivePoints(object, glass_data_file, fname);
  _glassLoadComponentsAndStructure(&component_list, &structure_tree, glass_data_file, fname);

  fclose(glass_data_file);

  /* Convert the temp data into a glass object */
  _glassGenerateComponentTree(object,
			      component_list,
			      structure_tree);

  /* Clean out all the temporary data */
  _glassFreeComponentList(component_list);
  _glassFreeStructureTree(structure_tree);

  /* Build some arrays that link to the variables and the triangle lists for fast access */
  _glassObjectBuildArrays(object);

  /* Convert the triangle lists into openGL display lists if possible */
  _glassObjectMakeDisplayLists(object);

  /* Work out which components are required for active point calculations */
  _glassComponentTestActivePointr(object->child);

  /* Finally! The glass object in all it's glory -- hopefully */
  return object;
}

/* Go through the file and load only the object info, i.e. anything
   that's not within a begin-end pair */
void _glassLoadObjectInfo(GlassObject *object, FILE *glass_data_file,
			  const char *fname)
{
  char *word = NULL, *data;
  int nbegins = 0;

  rewind(glass_data_file);

  do
  {
    word = _getNextWord(glass_data_file);

    if(word)
    {
      /* If inside begin-end pairs */
      if(nbegins > 0)
      {
	if(!strcmp(word, "end"))
	{
	  nbegins--;
	  if(nbegins < 0)
	  {
	    nbegins = 0;
	    fprintf(stderr,
		    "Glass Warning: Model file contains an `end' with no corresponding `begin'\n");
	  }
	}
      }
      else /* Otherwise check for root info */
      {
	data = _getNextWord(glass_data_file);
	if(data != NULL)
	{
	  if(!strcmp("name", word))
	    _glassObjectSetName(object, data);
	  else if(!strcmp("version", word))
	  {
	    char *glass_version = glassGetVersion();
	    if(strcmp(glass_version, data))
	    {
	      fprintf(stderr,
		      "Glass Warning: Model file has version \"%s\", "
		      "whereas glass library is version \"%s\"\n", data, glass_version);
	    }
	  }
	  else if(!strcmp(word, "begin"))
	    nbegins++;

	  free(data);
	}
      }
    }
  } while(word != NULL);
}

/* Go through the file and load only the textures */
void _glassLoadTextures(GlassObject *object, FILE *glass_data_file,
			const char *fname)
{
  char *word = NULL;

  rewind(glass_data_file);

  do
  {
    word = _getNextWord(glass_data_file);

    if(word == NULL);
    else if(!strcmp(word, "begin"))
    {
      free(word);
      word = _getNextWord(glass_data_file);
      if(!strcmp(word, "texture"))
	if(!_glassLoadTexture(object, glass_data_file, fname))
	  fprintf(stderr,
		  "Glass Error: Error loading texture in file \"%s\"\n",
		  fname);
    }

  } while(word != NULL);
}

/* Go through the file and load only the materials */
void _glassLoadMaterials(GlassObject *object, FILE *glass_data_file,
			 const char *fname)
{
  char *word = NULL;

  rewind(glass_data_file);

  do
  {
    word = _getNextWord(glass_data_file);

    if(word == NULL);
    else if(!strcmp(word, "begin"))
    {
      free(word);
      word = _getNextWord(glass_data_file);
      if(!strcmp(word, "material"))
	if(!_glassLoadMaterial(object, glass_data_file))
	  fprintf(stderr,
		  "Glass Error: Error loading material in file \"%s\"\n",
		  fname);
    }
  } while(word != NULL);
}

/* Go through the file and load only the triangle lists */
void _glassLoadTriangleLists(GlassObject *object, FILE *glass_data_file,
			     const char *fname)
{
  char *word = NULL;

  rewind(glass_data_file);

  do
  {
    word = _getNextWord(glass_data_file);

    if(word == NULL);
    else if(!strcmp(word, "begin"))
    {
      free(word);
      word = _getNextWord(glass_data_file);
      if(!strcmp(word, "triangle_list"))
	if(!_glassLoadTriangleList(object, glass_data_file))
	  fprintf(stderr,
		  "Glass Error: Error loading triangle list in file \"%s\"\n",
		  fname);
    }
  } while(word != NULL);
}

/* Go through the file and load only the variables */
void _glassLoadVariables(GlassObject *object, FILE *glass_data_file,
			 const char *fname)
{
  char *word = NULL;

  rewind(glass_data_file);

  do
  {
    word = _getNextWord(glass_data_file);

    if(word == NULL);
    else if(!strcmp(word, "begin"))
    {
      free(word);
      word = _getNextWord(glass_data_file);
      if(!strcmp(word, "variable"))
	if(!_glassLoadVariable(object, glass_data_file))
	  fprintf(stderr,
		  "Glass Error: Error loading variable in file \"%s\"\n",
		  fname);
    }
  } while(word != NULL);
}

/* Go through the file and load only the active points */
void _glassLoadActivePoints(GlassObject *object, FILE *glass_data_file,
			    const char *fname)
{
  char *word = NULL;

  rewind(glass_data_file);

  do
  {
    word = _getNextWord(glass_data_file);

    if(word == NULL);
    else if(!strcmp(word, "begin"))
    {
      free(word);
      word = _getNextWord(glass_data_file);
      if(!strcmp(word, "active_point"))
	if(!_glassLoadActivePoint(object, glass_data_file))
	  fprintf(stderr,
		  "Glass Error: Error loading active point in file \"%s\"\n",
		  fname);
    }
  } while(word != NULL);
}

/* Go through the file and load only the components, and structure info */
void _glassLoadComponentsAndStructure(GlassComponentList **component_list,
				      GlassStructureTree **structure_tree,
				      FILE *glass_data_file,
				      const char *fname)
{
  char *word = NULL;

  rewind(glass_data_file);

  do
  {
    word = _getNextWord(glass_data_file);

    if(word == NULL);
    else if(!strcmp(word, "begin"))
    {
      free(word);
      word = _getNextWord(glass_data_file);

      if(!strcmp("component", word))
      {
	if(!_glassLoadComponent(component_list, glass_data_file))
	    fprintf(stderr,
		    "Glass Error: Error loading component in file \"%s\"\n",
		    fname);
      }

      else if(!strcmp("structure", word))
      {
	if(!_glassLoadStructure(structure_tree, glass_data_file))
	  fprintf(stderr,
		  "Glass Error: Error loading component structure in file \"%s\"\n",
		  fname);
      }
    }
  } while(word != NULL);
}

/* Makes an entry for a component list -- used to load the raw component data from a file */
GlassComponentList *_glassComponentListNew(void)
{
  GlassComponentList *list = malloc(sizeof(GlassComponentList));
  int i;

  list->index = -1;
  list->next_index = -1;
  list->child_index = -1;

  list->list_name = NULL;
  list->apoint_name = NULL;
  for(i = 0; i < 4; i++)
    list->variable_names[i] = NULL;
  list->component = NULL;
  list->next = NULL;

  return list;
}

/* Makes an entry for a structure tree -- used to load the raw structure from a file */
GlassStructureTree *_glassStructureTreeNew(void)
{
  GlassStructureTree *structure = malloc(sizeof(GlassStructureTree));

  structure->index = -1;
  structure->component = NULL;
  structure->child = NULL;
  structure->next = NULL;

  return structure;
}

/* Returns a word from the file, ie removes whitespace and ignores comments (#) */
char *_getNextWord(FILE *file)
#define MAX_WORD_LENGTH 50
{
  char word[MAX_WORD_LENGTH + 1];
  int i = 0, input_char, is_word;

  do {
    /* Clear whitespace */
    do
      input_char = fgetc(file);
    while(isspace(input_char));
    is_word = 1;

    /* Remove comments */
    if(input_char == '#')
    {
      while((input_char != EOF) && (input_char != '\n'))
	input_char = fgetc(file);
      is_word = 0;
    }
  } while(!is_word); /* Repeat until whitespace has been cleared and no comment is present */

  if(input_char == EOF)
    return NULL;

  /* Get words (eg "Eat me") */
  if(input_char == '\"')
  {
    do
    {
      input_char = fgetc(file);
      word[i] = (char) input_char;
      i++;
    } while((input_char != '\"') && (input_char != EOF) && (i < MAX_WORD_LENGTH));
    word[i-1] = '\0';
  }
  else /* Get a word surrounded by whitespace */
  {
    while(!isspace(input_char) && (input_char != EOF) && (i < MAX_WORD_LENGTH))
    {
      word[i] = (char) input_char;
      input_char = fgetc(file);
      i++;
    }
    word[i] = '\0';
  }

  return strdup(word);
}

/* Gets the next string from the file and assumes a GLfloat which it converts to and returns */
GLfloat _getNextFloat(FILE *glass_data_file)
{
  char *float_string = _getNextWord(glass_data_file);
  GLfloat output = (GLfloat) atof(float_string);

  if(float_string == NULL)
  {
    free(float_string);
    return FLOAT_ERROR;
  }
  free(float_string);

  return output;
}

/* Load a texture image */
int _glassLoadTexture(GlassObject *object, FILE *glass_data_file, const char *fname)
#define MAX_NAME_LENGTH 128
{
  GlassTexture *texture = _glassTextureNew();
  char *string = NULL, *data = NULL;
  char *dir, texture_file[MAX_NAME_LENGTH];
  int i, l;

  /* Error checking */
  if(!texture)
    return FALSE;

  /* Add the texture into the glass object */
  _glassObjectAddTexture(object, texture);

  /* Get the properties of the texture */
  do
  {
    free(string);
    free(data);
    if(!(string = _getNextWord(glass_data_file)))
      return FALSE;
    if(strcmp(string, "end"))
    {
      data = _getNextWord(glass_data_file);
      if(!data)
      {
	free(string);
	return FALSE;
      }
    }
    else
      data = NULL;

    if(!strcmp("name", string))
      _glassTextureSetName(texture, data);

    else if(!strcmp("image", string))
    {
      /* Get the directory that the glass model is in */
      dir = strdup(fname);
      l = strlen(dir);
      for(i = l-2; i >= 0; i--)
      {
	if(dir[i] == '/') /* !!note change to '\' for win32 */
	{
	  dir[i+1] = '\0';
	  break;
	}
      }
      if(i < 0)
	_glassTextureLoadRGB(texture, data);
      else
      {
	snprintf(texture_file, MAX_NAME_LENGTH, "%s%s", dir, data);
	_glassTextureLoadRGB(texture, texture_file);
      }
    }

    else if(string  && strcmp("end", string))
      fprintf(stderr,
	      "Glass Warning: Unknown option \"%s\" in texture\n",
	      string);
  } while(string && strcmp("end", string));

  free(string);
  free(data);

  if(!_glassTextureGetName(texture))
    fprintf(stderr, "Glass Warning: Texture without an name\n");

  return TRUE;
}

/* Load a material */
int _glassLoadMaterial(GlassObject *object, FILE *glass_data_file)
{
  GlassMaterial *material = _glassMaterialNew();
  char *string = NULL, *data;
  GLfloat col[4];
  int i;

  /* Error checking */
  if(material == NULL)
    return FALSE;

  /* Add the material into the glass object */
  _glassObjectAddMaterial(object, material);

  /* Get the properties of the material */
  do
  {
    free(string);
    if((string = _getNextWord(glass_data_file)) == NULL)
      return FALSE;

    if(!strcmp("name", string))
    {
      data = _getNextWord(glass_data_file);
      _glassMaterialSetName(material, data);
      free(data);
    }

    else if(!strcmp("texture", string))
    {
      data = _getNextWord(glass_data_file);
      _glassMaterialSetTexture(material, _glassObjectFindTexture(object, data));
      if(_glassMaterialGetTexture(material) == NULL)
      {
	fprintf(stderr, "Glass Warning: Texture \"%s\" not found \n", data);
      }
      free(data);
    }

    else if(!strcmp("ambient", string))
    {
      for(i = 0; i < 4; i++)
	col[i] = _getNextFloat(glass_data_file);
      _glassMaterialSetAmbientv(material, col);
    }

    else if(!strcmp("diffuse", string))
    {
      for(i = 0; i < 4; i++)
	col[i] = _getNextFloat(glass_data_file);
      _glassMaterialSetDiffusev(material, col);
    }

    else if(!strcmp("specular", string))
    {
      for(i = 0; i < 4; i++)
	col[i] = _getNextFloat(glass_data_file);
      _glassMaterialSetSpecularv(material, col);
    }

    else if(!strcmp("emission", string))
    {
      for(i = 0; i < 4; i++)
	col[i] = _getNextFloat(glass_data_file);
      _glassMaterialSetEmissionv(material, col);
    }

    else if(!strcmp("shininess", string))
      _glassMaterialSetShininess(material, _getNextFloat(glass_data_file));

    else if((string != NULL) && (strcmp("end", string)))
      fprintf(stderr,
	      "Glass Warning: Unknown option \"%s\" in material\n",
	      string);
  } while((string != NULL) && (strcmp("end", string)));

  free(string);

  if(_glassMaterialGetName(material) == NULL)
    fprintf(stderr, "Glass Warning: Material without an name\n");

  return TRUE;
}

/* Loads a triangle list by loading triangles until hits "end" */
int _glassLoadTriangleList(GlassObject *object, FILE *glass_data_file)
{
  GlassTriangleList *tlist = _glassTriangleListNew("");
  char *string = NULL;

  /* Error checking */
  if(tlist == NULL)
    return FALSE;

  /* Add the triangle list into the glass object */
  _glassObjectAddList(object, tlist);

  /* Get the properties of the triangle list */
  do
  {
    free(string);
    if((string = _getNextWord(glass_data_file)) == NULL)
      return FALSE;

    if(!strcmp("name", string))
    {
      free(string);
      if((string = _getNextWord(glass_data_file)) == NULL)
	return FALSE;
      _glassTriangleListSetName(tlist, string);
    }

    else if(!strcmp("{", string))
      _glassLoadTriangle(object, tlist, glass_data_file);

    else if((string != NULL) && (strcmp("end", string)))
      fprintf(stderr,
	      "Glass Warning: Unknown option \"%s\" in triangle_list\n",
	      string);
  } while((string != NULL) && (strcmp("end", string)));

  free(string);

  if(_glassTriangleListGetName(tlist) == NULL)
    fprintf(stderr, "Glass Warning: Triangle List without an name\n");

  return TRUE;
}

/* Loads one triangle from a file */
int _glassLoadTriangle(GlassObject *object, GlassTriangleList *list, FILE *glass_data_file)
{
  GlassTriangle *triangle;
  GlassMaterial *material = NULL;
  char *string = NULL, *tmp_string = NULL;
  GLfloat pos[3][3], col[3][4], nor[3][3], tex[3][2];
  int i, j, is_alpha = FALSE, float_error;

  /* Set the defaults */
  for(i = 0; i < 3; i++)
  {
    for(j = 0; j < 3; j++)
    {
      pos[i][j] = 0.0;
      col[i][j] = 1.0;
      nor[i][j] = 0.0;
    }
    col[i][3] = 1.0;
    nor[i][2] = 1.0;
    tex[i][0] = 0.0;
    tex[i][1] = 0.0;
  }

  /* Load the availiable data */
  do {
    free(string);
    free(tmp_string);

    string = _getNextWord(glass_data_file);
    if(string == NULL)
      return FALSE;
    float_error = FALSE;

    if(!strcmp("material", string))
    {
      tmp_string = _getNextWord(glass_data_file);
      material = _glassObjectFindMaterial(object, tmp_string);
    }

    else if(!strcmp("pos", string))
    {
      tmp_string = _getNextWord(glass_data_file);
      j = atoi(tmp_string);

      if(j >= 0 && j < 3)
	for(i = 0; i < 3; i++)
	  if((pos[j][i] = _getNextFloat(glass_data_file)) == FLOAT_ERROR)
	    float_error = TRUE;
    }

    else if(!strcmp("col", string))
    {
      tmp_string = _getNextWord(glass_data_file);
      j = atoi(tmp_string);

      for(i = 0; i < 3; i++)
	if((col[j][i] = _getNextFloat(glass_data_file)) == FLOAT_ERROR)
	  float_error = TRUE;
    }

    else if(!strcmp("alpha", string))
    {
      tmp_string = _getNextWord(glass_data_file);
      j = atoi(tmp_string);

      if((col[j][3] = _getNextFloat(glass_data_file)) == FLOAT_ERROR)
	float_error = TRUE;
      is_alpha = TRUE;
    }

    else if(!strcmp("nor", string))
    {
      tmp_string = _getNextWord(glass_data_file);
      j = atoi(tmp_string);

      for(i = 0; i < 3; i++)
	if((nor[j][i] = _getNextFloat(glass_data_file)) == FLOAT_ERROR)
	  float_error = TRUE;
    }

    else if(!strcmp("tex", string))
    {
      tmp_string = _getNextWord(glass_data_file);
      j = atoi(tmp_string);

      for(i = 0; i < 2; i++)
	if((tex[j][i] = _getNextFloat(glass_data_file)) == FLOAT_ERROR)
	  float_error = TRUE;
    }

    else if(strcmp("}", string))
      fprintf(stderr, "Glass Warning: Unknown option %s found in triangle definition\n", string);
  } while(!float_error && strcmp("}", string));

  if(string[0] != '}')
    free(tmp_string);
  free(string);

  /* If some of the values weren't found */
  if(float_error)
    fprintf(stderr, "Glass Warning: Incomplete triangle found\n");

  /* Make a triangle, and copy the data */
  if(is_alpha)
    triangle = _glassAlphaTriangleNew();
  else
    triangle = _glassTriangleNew();

  _glassTriangleSetMaterial(triangle, material);

  /* Data */
  for(i = 0; i < 3; i++)
  {
    for(j = 0; j < 3; j++)
    {
      triangle->pos[i][j] = pos[i][j];
      triangle->col[i][j] = col[i][j];
      triangle->nor[i][j] = nor[i][j];
    }
    if(is_alpha)
      triangle->col[i][3] = col[i][3];
    triangle->tex[i][0] = tex[i][0];
    triangle->tex[i][1] = tex[i][1];
  }

  /* Attach triangle to list */
  _glassTriangleListAddTriangle(list, triangle);

  return TRUE;
}

/* Loads a variable */
int _glassLoadVariable(GlassObject *object, FILE *glass_data_file)
{
  GlassVariable *variable = _glassVariableNew("", 0.0, 1.0, 0.0);
  char *option = NULL, *data = NULL;

  /* Error checking */
  if(variable == NULL)
    return FALSE;

  /* Add the variable into the glass object */
  _glassObjectAddVariable(object, variable);

  do
  {
    free(option);
    free(data);

    option = _getNextWord(glass_data_file);
    if(option  && strcmp("end", option))
      data = _getNextWord(glass_data_file);

    /* Error checking */
    if((option == NULL) || (data == NULL))
    {
      free(option);
      free(data);
      return FALSE;
    }

    if(!strcmp("name", option))
      _glassVariableSetName(variable, data);
    else if(!strcmp("min", option))
      variable->min = (GLfloat) atof(data);
    else if(!strcmp("max", option))
      variable->max = (GLfloat) atof(data);
    else if(!strcmp("value", option))
      variable->value = (GLfloat) atof(data);
    else if(strcmp("end", option))
      fprintf(stderr,
	      "Glass Warning: Unknown option %s (with data \"%s\"), in variable\n",
	      option,
	      data);

  } while(strcmp("end", option));

  free(option);

  if(_glassVariableGetName(variable) == NULL)
    fprintf(stderr, "Glass Warning: Variable without an name\n");

  return TRUE;
}

/* Loads an active point */
int _glassLoadActivePoint(GlassObject *object, FILE *glass_data_file)
{
  GlassActivePoint *apoint = _glassActivePointNew();
  char *word = NULL, *name = NULL;
  GLfloat idir[3];
  int i;

  /* Error checking */
  if(apoint == NULL)
    return FALSE;

  /* Add the active point into the glass object */
  _glassObjectAddActivePoint(object, apoint);

  do
  {
    free(word);
    if((word = _getNextWord(glass_data_file)) == NULL)
      return FALSE;

    if(!strcmp("name", word))
    {
      name = _getNextWord(glass_data_file);
      _glassActivePointSetName(apoint, name);
      free(name);
    }
    else if(!strcmp("initial_dir", word))
    {
      for(i = 0; i < 3; i++)
	idir[i] = _getNextFloat(glass_data_file);
      _glassActivePointSetInitDirv(apoint, idir);
    }
    else if(strcmp("end", word))
      fprintf(stderr,
	      "Glass Warning: Unknown option %s, in active point\n",
	      word);

  } while(strcmp("end", word));

  free(word);

  if(_glassActivePointGetName(apoint) == NULL)
    fprintf(stderr, "Glass Warning: Active point without an name\n");

  return TRUE;
}

/* Loads a component */
int _glassLoadComponent(GlassComponentList **component_list, FILE *glass_data_file)
{
  GlassComponentList *entry = _glassComponentListNew(), **list_ptr = component_list;
  GlassComponent *component = _glassComponentNew();
  char *option = NULL, *data = NULL;

  /* Error checking */
  if((entry == NULL) || (component == NULL))
  {
    fprintf(stderr, "Glass Error: Unable to make space for new component\n");
    free(entry);
    free(component);
    return FALSE;
  }

  /* Find the end of the list */
  while(*list_ptr != NULL)
    list_ptr = &(*list_ptr)->next;

  /* Attach new entry to the list */
  *list_ptr = entry;
  /* Attach the new component to the component list entry */
  entry->component = component;

  /* Check the data until an end */
  do {
    /* Input from the file */
    free(option);
    free(data);

    option = _getNextWord(glass_data_file);
    if((option != NULL) && (strcmp("end", option)))
       data = _getNextWord(glass_data_file);
    if((option == NULL) || (data == NULL))
    {
      free(option);
      free(data);
      return FALSE;
    }

    /* Deal with the input */
    if(!strcmp("index", option))
      entry->index = atoi(data);
    else if(!strcmp("next_index", option))
      entry->next_index = atoi(data);
    else if(!strcmp("child_index", option))
      entry->child_index = atoi(data);
    else if(!strcmp("list", option))
    {
      free(entry->list_name);
      entry->list_name = strdup(data);
    }
    else if(!strcmp("active_point", option))
    {
      free(entry->apoint_name);
      entry->apoint_name = strdup(data);
    }
    else if(!strcmp("transform", option))
      _glassLoadTransform(&data, entry, glass_data_file);

    else if(strcmp("end", option))
      fprintf(stderr,
	      "Glass Warning: Unknown option \"%s\" (with data \"%s\"), in component\n",
	      option,
	      data);
  } while(strcmp("end", option));

  free(option);

  return TRUE;
}

/* Load a transform of type specified by *type */
int _glassLoadTransform(char **data, GlassComponentList *entry, FILE *glass_data_file)
{
  GlassTransformType type = _glassConvertTransform(*data);
  GlassTransform *transform = _glassComponentGetTransform(entry->component);
  int nparams, i;

  _glassTransformSetType(transform, type);

  /* Amount of floats used by the transform */
  nparams = _glassGetnParams(type);

  for(i = 0; i < nparams; i++)
  {
    /* Get the number */
    free(*data);
    *data = _getNextWord(glass_data_file);

    if(*data == NULL)
      return FALSE;

    /* If is a variable */
    if(!strncmp(*data, "V::", 3))
    {
      free(entry->variable_names[i]);
      entry->variable_names[i] = strdup(*data+3);
    }
    else /* Otherwise is a constant */
    {
      free(entry->variable_names[i]);
      entry->variable_names[i] = NULL;
      *transform->values[i] = atof(*data);
    }
  }

  return TRUE;
}

/* Convert a string to a transform type */
GlassTransformType _glassConvertTransform(const char *word)
#define N_TRANSFORMS 14
{
  char transforms[][20] = {"TRANSLATE", "ROTATE", "SCALE", "TRANSLATE_X",
			   "TRANSLATE_Y", "TRANSLATE_Z", "ROTATE_X", "ROTATE_Y",
			   "ROTATE_Z", "SCALE_X", "SCALE_Y", "SCALE_Z", "SCALE_EVEN",
			   "NO_TRANSFORM"};
  int i;

  for(i = 0; i < N_TRANSFORMS; i++)
    if(!strcmp(transforms[i], word))
      switch(i)
      {
      case 0: /* "TRANSLATE" */
	return GLASS_TRANSLATE;
	break;
      case 1: /* "ROTATE" */
	return GLASS_ROTATE;
	break;
      case 2: /* "SCALE" */
	return GLASS_SCALE;
	break;
      case 3: /* "TRANSLATE_X" */
	return GLASS_TRANSLATE_X;
	break;
      case 4: /* "TRANSLATE_Y" */
	return GLASS_TRANSLATE_Y;
	break;
      case 5: /* "TRANSLATE_Z" */
	return GLASS_TRANSLATE_Z;
	break;
      case 6: /* "ROTATE_X" */
	return GLASS_ROTATE_X;
	break;
      case 7: /* "ROTATE_Y" */
	return GLASS_ROTATE_Y;
	break;
      case 8: /* "ROTATE_Z" */
	return GLASS_ROTATE_Z;
	break;
      case 9: /* "SCALE_X" */
	return GLASS_SCALE_X;
	break;
      case 10: /* "SCALE_Y" */
	return GLASS_SCALE_Y;
	break;
      case 11: /* "SCALE_Z" */
	return GLASS_SCALE_Z;
	break;
      case 12: /* "SCALE_EVEN" */
	return GLASS_SCALE_EVEN;
	break;
      case 13: /* "NO_TRANSFORM" */
	return GLASS_NONE;
	break;
      }

  fprintf(stderr, "Glass Warning: Unknown transform type (%s)\n", word);
  return GLASS_NONE;
}

struct _LineList
{
  char *line;
  struct _LineList *next;
};

/* Loads the structure of the glass object */
int _glassLoadStructure(GlassStructureTree **structure_tree,
			FILE *glass_data_file)
#define MAX_LINE_LENGTH 100
{
  struct _LineList *line_list = malloc(sizeof(struct _LineList)), *line_ptr = line_list, *tmp_ptr;
  char **input_array = NULL;
  char input_line[MAX_LINE_LENGTH];
  int num = 0, width = 0, i, j;

  /* Error Checking */
  if(line_list == NULL)
    return FALSE;

  /* Cap end */
  line_list->next = NULL;

  /* Copy the data into the _LineList structure */
  do
  {
    /* Get one line from the file */
    fgets(input_line, MAX_LINE_LENGTH, glass_data_file);
    /* Replace the newline with a '\0' */
    strtok(input_line, "\n");

    if((input_line != NULL) && strcmp("end", input_line))
    {
      line_ptr->line = strdup(input_line);

      line_ptr->next = malloc(sizeof(struct _LineList));

      line_ptr = line_ptr->next;

      if(strlen(input_line) > width)
	width = strlen(input_line);
      num++;
      }
    } while((input_line != NULL) && strcmp("end", input_line));

  /* Note: the last line isn't actually used */
  line_ptr->line = NULL;
  line_ptr->next= NULL; /* cap end */

  /* Reset the line list pointer */
  line_ptr = line_list;

  /* Now convert the linked list to a 2d array */
  input_array = malloc(sizeof(char *) * num);
  for(i = 0; i < num; i++)
  {
    /* Make room for the max size line */
    input_array[i] = malloc(sizeof(char) * width);
    /* Set the entries to ' ' -- this is so the aray can be assumed to be square even though
       some entries will be longer than others (fills the otherwise unused spaces) */
    for(j = 0; j < width; j++)
      input_array[i][j] = ' ';
    /* Copy the linked list entry */
    strncpy(input_array[i], line_ptr->line, strlen(line_ptr->line));
    line_ptr = line_ptr->next;
  }

  /* Free the old linked list */
  line_ptr = line_list;
  while(line_ptr != NULL)
  {
    tmp_ptr = line_ptr->next;
    free(line_ptr->line);
    free(line_ptr);
    line_ptr = tmp_ptr;
  }

  /* Phew! So now we have a 2d array of width "width" and height "num" containing a
     text based tree, next put that into a structure */

  /* Recursively gets the relatives*/
  _getRelatives(input_array, structure_tree, 0, 0, width, num);

  /* Free the 2d array */
  for(i = 0; i < num; i++)
    free(input_array[i]);
  free(input_array);

  /* So now we *should* have a data structure containig the relationship between the indicies...
     We should. Hopefully. */
  
  return TRUE;
}

/* Recursively gets the tree of component indicies */
int _getRelatives(char **input_array,
		  GlassStructureTree **structure,
		  int x, int y,
		  int width, int height)
{
  int index = _getInt(input_array, x, y, width, height);

  /* If isn't an entry */
  if(index == -1)
    return FALSE;

  /* If is moving sideways eg 001-----002 will have to move right twice for sibling of 001
     but should skip the middle part ("---") */
  if(index == -2)
  {
    _getRelatives(input_array, structure, x + 1, y, width, height);
    return TRUE;
  }

  /* Make a new structure */
  *structure = _glassStructureTreeNew();
  (*structure)->index = _getInt(input_array, x, y, width, height);

  /* Get the siblings */
  _getRelatives(input_array, &(*structure)->next, x + 1, y, width, height);

  /* Get the children */
  _getRelatives(input_array, &(*structure)->child, x, y + 1, width, height);

  return TRUE;
}

/* Gets a number from the array generated above */
int _getInt(char **array, int x, int y, int width, int height)
{
  char number[4];

  /* Check bounds */
  if((x > width / 4) || (y >= height))
    return -1;

  /* Get a 3 digit "number" from the data array */
  number[0] = array[y][x * 4];
  number[1] = array[y][x * 4 + 1];
  number[2] = array[y][x * 4 + 2];
  number[3] = '\0';

  /* If isn't an entry */
  if(!strcmp("   ", number))
    return -1;

  /* If it links to an entry further over return -2 */
  if(!strcmp("---", number))
    return -2;

  /* The index of the component */
  return atoi(number);
}

/* Generates a glass object out of the components, variables, and triangles lists based on their
   names/indicies */
int _glassGenerateComponentTree(GlassObject *object,
				GlassComponentList *component_list,
				GlassStructureTree *structure_tree)
{
  GlassVariable *variable;
  GlassComponentList *com_list_ptr = component_list;
  GlassTransform *transform;
  int i, nparams;

  /* Link the components with variables to the correct values and link the components to the
     correct triangle lists */
  while(com_list_ptr != NULL)
  {
    transform = com_list_ptr->component->transform;

    /* Connect the transform to its variables */
    if(transform != NULL)
      nparams = _glassGetnParams(transform->type);
    else
      nparams = -1;

    for(i = 0; i < nparams; i++)
    {
      /* If this parameter is a variable */
      if(com_list_ptr->variable_names[i])
      {
	variable = _glassObjectFindVariable(object, com_list_ptr->variable_names[i]);
	if(variable)
	  _glassTransformSetVariable(transform, i, variable);
	else
	{
	  fprintf(stderr, "Glass Warning: Variable \"%s\" not found\n",
		  com_list_ptr->variable_names[i]);
	  _glassTransformSetConstant(transform, i, 0.0);
	}
      }
    }

    /* If has an associated triangle list */
    if(com_list_ptr->list_name)
    {
      com_list_ptr->component->list = _glassObjectFindList(object, com_list_ptr->list_name);
      if(_glassComponentGetList(com_list_ptr->component) == NULL)
	fprintf(stderr, "Glass Warning: Triangle List \"%s\" not found\n",
		com_list_ptr->list_name);
    }

    /* If has an associated active point */
    if(com_list_ptr->apoint_name)
    {
      com_list_ptr->component->apoint = _glassObjectFindActivePoint(object,
								    com_list_ptr->apoint_name);
      if(_glassComponentGetActivePoint(com_list_ptr->component) == NULL)
	fprintf(stderr, "Glass Warning: Active point \"%s\" not found\n",
		com_list_ptr->apoint_name);
    }

    /* Move on to the next component */
    com_list_ptr = com_list_ptr->next;
  }
  
  /* Now build the component tree */
  /* This is done one of two ways,
     a) If a structure tree was supplied, they are linked together in this format.
        This is the origional way it was done, and is easiest if you're writing the
	file yourself.
     b) If a structure tree wasn't supplied, use the next_index, and child_index entries.
        This is the way glass saves it's files, mainly because I was having trouble
        making the _glassWriteStructure() function. */
  /* Recursively link the components to eachother */
  if(component_list != NULL)
  {
    if(structure_tree == NULL)
      object->child = _linkComponents(component_list);
    else
    {
      _linkComponentStructure(structure_tree, component_list);
      /* Now the structure contains the first component which is linked into the tree, so...
	 the object just has to link to this and then it'll have the whole tree */
      object->child = structure_tree->component;
    }
  }

  return TRUE;
}

/* Link the components together using the next_ and child_ indicies */
GlassComponent *_linkComponents(GlassComponentList *component_list)
{
  GlassComponentList *list_ptr = component_list, *list;

  /* Go through each component */
  while(list_ptr != NULL)
  {
    /* Link to its children */
    if(list_ptr->child_index != -1)
    {
      /* Find the component with the index child_index */
      list = _glassFindComponentList(component_list, list_ptr->child_index);
      if(list == NULL)
	break;
      list_ptr->component->child = list->component;
    }

    /* Link to its siblings */
    if(list_ptr->next_index != -1)
    {
      /* Find the component with the index next_index */
      list = _glassFindComponentList(component_list, list_ptr->next_index);
      if(list == NULL)
	break;
      list_ptr->component->next = list->component;
    }

    list_ptr = list_ptr->next;
  }

  /* Find the component with index 0 -- the root component */
  list = _glassFindComponentList(component_list, 0);

  if(list == NULL)
    return NULL;
  else
    return list->component;
}

/* Recursively link the components together as it says in the structure tree */
int _linkComponentStructure(GlassStructureTree *structure, GlassComponentList *component_list)
{
  GlassComponentList *list = _glassFindComponentList(component_list, structure->index);

  /* Find the component for the tree nodes index */
  if(list->component == NULL)
    return FALSE;
  else
    structure->component = list->component;

  /* Do children */
  if(structure->child != NULL)
  {
    _linkComponentStructure(structure->child, component_list);
    structure->component->child = structure->child->component;
  }

  /* Do siblings */
  if(structure->next != NULL)
  {
    _linkComponentStructure(structure->next, component_list);
    structure->component->next = structure->next->component;
  }

  return TRUE;
}

/* Find a list from a component list with the index, index */
GlassComponentList *_glassFindComponentList(GlassComponentList *list, int index)
{
  while((list != NULL) && (list->index != index))
    list = list->next;

  if(list == NULL)
  {
    fprintf(stderr,
	    "Glass Warning: Unable to find component %d\n",
	    index);
    return NULL;
  }

  return list;
}

/* Free a component list -- doesn't free the components though */
int _glassFreeComponentList(GlassComponentList *component_list)
{
  GlassComponentList *list_ptr = component_list, *tmp;

  while(list_ptr != NULL)
  {
    tmp = list_ptr;
    list_ptr = list_ptr->next;

    free(tmp->list_name);
    free(tmp->apoint_name);
    free(tmp->variable_names[0]);
    free(tmp->variable_names[1]);
    free(tmp->variable_names[2]);
    free(tmp->variable_names[3]);
    free(tmp);
  }

  return TRUE;
}

/* Recursively frees a structure tree -- doesn't free the components though */
int _glassFreeStructureTree(GlassStructureTree *structure_tree)
{
  if(structure_tree == NULL)
    return TRUE;

  if(structure_tree->child != NULL)
    _glassFreeStructureTree(structure_tree->child);

  if(structure_tree->next != NULL)
    _glassFreeStructureTree(structure_tree->next);

  free(structure_tree);

  return TRUE;
}
