/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <malloc.h>
#include <string.h>

#include "glass.h"

/* Returns an empty triangle list */
GlassTriangleList *_glassTriangleListNew(const char *name)
{
  GlassTriangleList *tlist = malloc(sizeof(GlassTriangleList));

  if(tlist == NULL)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for triangle list\n");
    return NULL;
  }

  tlist->name = NULL;
  _glassTriangleListSetName(tlist, name);

  tlist->index = 0;
  tlist->alpha_index = 0;
  tlist->visible = TRUE;

  tlist->next = NULL;
  tlist->first = NULL;
  tlist->first_alpha = NULL;

  return tlist;
}

/* Set the name of the triangle list */
void _glassTriangleListSetName(GlassTriangleList *tlist, const char *name)
{
  free(tlist->name);
  if(name == NULL)
    tlist->name = strdup("");
  else
    tlist->name = strdup(name);
}

/* Get the name of the triangle list */
const char *_glassTriangleListGetName(GlassTriangleList *tlist)
{
  return tlist->name;
}

/* Add a triangle to the triangle list */
void _glassTriangleListAddTriangle(GlassTriangleList *tlist, GlassTriangle *triangle)
{
  GlassTriangle **tri_ptr;

  if(triangle->is_alpha)
    tri_ptr = &tlist->first_alpha;
  else
    tri_ptr = &tlist->first;

  /* Find the end of the list, and add the triangle */
  while(*tri_ptr != NULL)
    tri_ptr = &(*tri_ptr)->next;

  *tri_ptr = triangle;
}

/* Remove a triangle from a triangle list, return FALSE if triangle not in list */
int _glassTriangleListRemoveTriangle(GlassTriangleList *tlist, GlassTriangle *triangle)
{
  GlassTriangle *tri_ptr;

  if(triangle->is_alpha)
    tri_ptr = tlist->first_alpha;
  else
    tri_ptr = tlist->first;

  /* If no triangles */
  if(tri_ptr == NULL)
    return FALSE;

  /* If is the first triangle */
  if(tri_ptr == triangle)
  {
    if(triangle->is_alpha)
      tlist->first_alpha = triangle->next;
    else
      tlist->first = triangle->next;
    triangle->next = NULL;
    return TRUE;
  }

  /* Find the triangle before this one */
  while(tri_ptr != NULL && tri_ptr->next != triangle)
    tri_ptr = tri_ptr->next;

  /* If at the end of the list */
  if(tri_ptr == NULL)
    return FALSE;

  /* Skip out that triangle */
  tri_ptr->next = triangle->next;

  /* Unlink the triangle */
  triangle->next = NULL;

  return TRUE;
}

/* Try to make an openGL display list for the triangles, and one for the alpha triangles */
void _glassTriangleListMakeDisplayLists(GlassTriangleList *tlist)
{
  GLuint index = 0, alpha_index = 0;

  /* If has non-alpha triangles and doesn't yet have a display list */
  if(!tlist->index && tlist->first)
  {
    index = glGenLists(1);
    if(!index)
      return;

    glNewList(index, GL_COMPILE);
      _glassDrawTriangleList(tlist);
    glEndList();

    tlist->index = index;
  }

  /* If has alpha triangles and doesn't yet have a display list */
  if(!tlist->alpha_index && tlist->first_alpha)
  {
    alpha_index = glGenLists(1);
    if(!alpha_index)
      return;

    glNewList(alpha_index, GL_COMPILE);
      _glassDrawAlphaTriangleList(tlist);
    glEndList();

    tlist->alpha_index = alpha_index;
  }
}

/* Set the visiblility of the triangle list */
void _glassTriangleListSetVisibility(GlassTriangleList *tlist, int visible)
{
  tlist->visible = visible;
}

/* Free a triangle list */
void _glassFreeTriangleList(GlassTriangleList *tlist)
{
  free(tlist->name);
  free(tlist);
}

/* Free a triangle list, and the triangles it uses */
void _glassFreeTriangleListChild(GlassTriangleList *tlist)
{
  GlassTriangle *tri_ptr = tlist->first, *tmp_ptr;

  /* Free the non-alpha triangles */
  while(tri_ptr != NULL)
  {
    tmp_ptr = tri_ptr->next;
    _glassFreeTriangle(tri_ptr);
    tri_ptr = tmp_ptr;
  }

  /* Free the alpha triangles */
  tri_ptr = tlist->first_alpha;
  while(tri_ptr != NULL)
  {
    tmp_ptr = tri_ptr->next;
    _glassFreeTriangle(tri_ptr);
    tri_ptr = tmp_ptr;
  }

  _glassFreeTriangleList(tlist);
}
