/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <malloc.h>
#include <math.h>

#include "glass.h"

/* Returns an empty triangle */
GlassTriangle *_glassTriangleNew(void)
{
  GlassTriangle *triangle = malloc(sizeof(GlassTriangle));
  int i, j;

  if(triangle == NULL)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for triangle\n");
    return NULL;
  }

  triangle->is_alpha = FALSE;

  triangle->pos = malloc(sizeof(GLfloat *) * 3);
  triangle->col = malloc(sizeof(GLfloat *) * 3);
  triangle->nor = malloc(sizeof(GLfloat *) * 3);
  triangle->tex = malloc(sizeof(GLfloat *) * 3);

  for(i = 0; i < 3; i++)
  {
    triangle->pos[i] = malloc(sizeof(GLfloat) * 3);
    triangle->col[i] = malloc(sizeof(GLfloat) * 3);
    triangle->nor[i] = malloc(sizeof(GLfloat) * 3);
    triangle->tex[i] = malloc(sizeof(GLfloat) * 2);
  }

  /* Set the defaults */
  for(i = 0; i < 3; i++)
  {
    for(j = 0; j < 3; j++)
    {
      triangle->pos[i][j] = 0.0;
      triangle->col[i][j] = 1.0;
      triangle->nor[i][j] = 0.0;
    }
    triangle->tex[i][0] = 0.0;
    triangle->tex[i][1] = 0.0;
    triangle->nor[i][2] = 1.0;
  }

  triangle->material = NULL;

  triangle->next = NULL;

  return triangle;
}

/* Returns an empty alpha triangle */
GlassTriangle *_glassAlphaTriangleNew(void)
{
  GlassTriangle *alpha_triangle = malloc(sizeof(GlassTriangle));
  int i, j;

  if(alpha_triangle == NULL)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for triangle\n");
    return NULL;
  }

  alpha_triangle->is_alpha = TRUE;

  alpha_triangle->pos = malloc(sizeof(GLfloat *) * 3);
  alpha_triangle->col = malloc(sizeof(GLfloat *) * 3);
  alpha_triangle->nor = malloc(sizeof(GLfloat *) * 3);
  alpha_triangle->tex = malloc(sizeof(GLfloat *) * 3);

  for(i = 0; i < 3; i++)
  {
    alpha_triangle->pos[i] = malloc(sizeof(GLfloat) * 3);
    alpha_triangle->col[i] = malloc(sizeof(GLfloat) * 4);
    alpha_triangle->nor[i] = malloc(sizeof(GLfloat) * 3);
    alpha_triangle->tex[i] = malloc(sizeof(GLfloat) * 2);
  }

  /* Set the defaults */
  for(i = 0; i < 3; i++)
  {
    for(j = 0; j < 3; j++)
    {
      alpha_triangle->pos[i][j] = 0.0;
      alpha_triangle->col[i][j] = 0.0;
      alpha_triangle->nor[i][j] = 0.0;
    }
    alpha_triangle->col[i][3] = 0.0;
    alpha_triangle->tex[i][0] = 0.0;
    alpha_triangle->tex[i][1] = 0.0;
    alpha_triangle->nor[i][2] = 1.0;
  }

  alpha_triangle->material = NULL;

  alpha_triangle->next = NULL;

  return alpha_triangle;
}

/* Set the position for a triangles vertex */
void _glassTriangleSetPosition(GlassTriangle *triangle, int vertex, GLfloat *pos)
{
  GLfloat *vpos = triangle->pos[vertex];

  vpos[0] = pos[0];
  vpos[1] = pos[1];
  vpos[2] = pos[2];
}

/* Get the position of a triangles vertex */
GLfloat *_glassTriangleGetPosition(GlassTriangle *triangle, int vertex)
{
  return triangle->pos[vertex];
}

/* Set the colour for a triangles vertex */
void _glassTriangleSetColour(GlassTriangle *triangle, int vertex, GLfloat *col)
{
  GLfloat *vcol = triangle->col[vertex];
  int i;

  vcol[0] = col[0];
  vcol[1] = col[1];
  vcol[2] = col[2];
  if(triangle->is_alpha)
    vcol[3] = col[3];

  /* Clip the values */
  for(i = 0; i < 3; i++)
    if(vcol[i] < 0.0)
      vcol[i] = 0.0;
    else if(vcol[i] > 1.0)
      vcol[i] = 1.0;

  if(triangle->is_alpha)
  {
    if(vcol[3] < 0.0)
      vcol[3] = 0.0;
    else if(vcol[3] > 1.0)
      vcol[3] = 1.0;
  }
}

/* Get the colour of a triangles vertex */
GLfloat *_glassTriangleGetColour(GlassTriangle *triangle, int vertex)
{
  return triangle->col[vertex];
}

/* Set the normal for a triangles vertex */
void _glassTriangleSetNormal(GlassTriangle *triangle, int vertex, GLfloat *nor)
{
  GLfloat *vnor = triangle->nor[vertex], len;

  len = sqrt(nor[0]*nor[0] + nor[1]*nor[1] + nor[2]*nor[2]);

  vnor[0] = nor[0] / len;
  vnor[1] = nor[1] / len;
  vnor[2] = nor[2] / len;
}

/* Get the normal of a triangles vertex */
GLfloat *_glassTriangleGetNormal(GlassTriangle *triangle, int vertex)
{
  return triangle->nor[vertex];
}

/* Set the texture co-ordinates for a triangles vertex */
void _glassTriangleSetTexCoord(GlassTriangle *triangle, int vertex, GLfloat *tex)
{
  GLfloat *vtex = triangle->tex[vertex];

  vtex[0] = tex[0];
  vtex[1] = tex[1];
}

/* Get the texture co-ordinate of a triangles vertex */
GLfloat *_glassTriangleGetTexCoord(GlassTriangle *triangle, int vertex)
{
  return triangle->tex[vertex];
}

/* Set the material of the triangle */
void _glassTriangleSetMaterial(GlassTriangle *triangle, GlassMaterial *material)
{
  if(!material)
  {
    triangle->material = NULL;
    return;
  }

  if(triangle->is_alpha != material->is_alpha)
    fprintf(stderr,
	    "Glass Warning: Alpha material attached to non-alpha triangle, or visa versa\n");

  triangle->material = material;
}

/* Get the material of a triangle */
GlassMaterial *_glassTriangleGetMaterial(GlassTriangle *triangle)
{
  return triangle->material;
}

/* Free a triangle */
void _glassFreeTriangle(GlassTriangle *triangle)
{
  int i;

  for(i = 0; i < 3; i++)
  {
    free(triangle->pos[i]);
    free(triangle->col[i]);
    free(triangle->nor[i]);
    free(triangle->tex[i]);
  }

  free(triangle->pos);
  free(triangle->col);
  free(triangle->nor);
  free(triangle->tex);
}
